-- Created by Giseldah

local p = _G; --p stands for package

-- ****************************** Start CalcStat ******************************

-- Floating point numbers bring errors into the calculation, both inside the Lotro-client and in this function collection. This is why a 100% match with the stats in Lotro is impossible.
-- Anyway, to compensate for some errors, we use a calculation deviation correction value. This makes for instance 24.49999999 round to 25, as it's assumed that 24.5 was intended as outcome of a formula.
DblCalcDev = 0.00000001;

function CalcStat(SName, SLvl, SParam)

	-- process parameters and parameter defaults

	if type(SName) ~= "string" then
		return "Missing stat name";
	end
	local SN = string.match(SName:upper(),"^%s*(-?%a+%d*%a+)%s*$"); -- allow nested digits in stat name and a starting -
	if SN == nil then
		SN = string.match(SName,"^%s*(-?%a+)%s*$"); -- to allow single character stat name
		if SN == nil then
			return "Illegal stat name";
		end
	end
	SN = string.upper(SN);

	local L = 1; -- default L
	if type(SLvl) == "number" then
		L = SLvl;
	elseif type(SLvl) ~= "nil" then
		return "Illegal level";
	end
	local Lm = L-DblCalcDev;
	local Lp = L+DblCalcDev;

	local N = 1; -- default N
	local C = ""; -- default C
	if type(SParam) == "number" then
		N = SParam;
	elseif type(SParam) == "string" then
		C = SParam;
	elseif type(SParam) ~= "nil" then
		return "Illegal N or C";
	end

	-- binary search tree (generated code)

	if SN < "PARRYPRATP" then
		if SN < "EVADEPRATPCAP" then
			if SN < "CHAMPIONCDARMOURTYPE" then
				if SN < "BPEPRATPCAP" then
					if SN < "BLOCKPRATPB" then
						if SN < "BEORNINGCDCANBLOCK" then
							if SN < "ARMOURPENT" then
								if SN == "-VERSION" then
									return "2.2.3p";
								else
									return 0;
								end
							elseif SN > "ARMOURPENT" then
								if SN > "ARMOURPENTADJ" then
									if SN == "BEORNINGCDARMOURTYPE" then
										return 3;
									else
										return 0;
									end
								elseif SN == "ARMOURPENTADJ" then
									if Lm <= 49 then
										return 1;
									elseif Lm <= 50 then
										return 0.8;
									else
										return 1;
									end
								else
									return 0;
								end
							else
								return StatLinInter("PntMPArmourPenT","TraitProg","ProgBArmour","ArmourPenTAdj",L,N,0);
							end
						elseif SN > "BEORNINGCDCANBLOCK" then
							if SN < "BLOCKPPRAT" then
								if SN == "BLOCKPBONUS" then
									return CalcStat("BPEPBonus",L);
								else
									return 0;
								end
							elseif SN > "BLOCKPPRAT" then
								if SN > "BLOCKPRATP" then
									if SN == "BLOCKPRATPA" then
										return CalcStat("BPEPRatPA",L);
									else
										return 0;
									end
								elseif SN == "BLOCKPRATP" then
									return CalcStat("BPEPRatP",L,N);
								else
									return 0;
								end
							else
								return CalcStat("BPEPPRat",L,N);
							end
						else
							if Lm <= 5 then
								return 0;
							else
								return 1;
							end
						end
					elseif SN > "BLOCKPRATPB" then
						if SN < "BPEPPRAT" then
							if SN < "BLOCKPRATPCAP" then
								if SN == "BLOCKPRATPC" then
									return CalcStat("BPEPRatPC",L);
								else
									return 0;
								end
							elseif SN > "BLOCKPRATPCAP" then
								if SN > "BLOCKPRATPCAPR" then
									if SN == "BPEPBONUS" then
										return 0;
									else
										return 0;
									end
								elseif SN == "BLOCKPRATPCAPR" then
									return CalcStat("BPEPRatPCapR",L);
								else
									return 0;
								end
							else
								return CalcStat("BPEPRatPCap",L);
							end
						elseif SN > "BPEPPRAT" then
							if SN < "BPEPRATPA" then
								if SN == "BPEPRATP" then
									return CalcPercAB(CalcStat("BPEPRatPA",L),CalcStat("BPEPRatPB",L),CalcStat("BPEPRatPCap",L),N);
								else
									return 0;
								end
							elseif SN > "BPEPRATPA" then
								if SN > "BPEPRATPB" then
									if SN == "BPEPRATPC" then
										return 0.5;
									else
										return 0;
									end
								elseif SN == "BPEPRATPB" then
									return CalcStat("BRatRounded",L,"BRatStandard");
								else
									return 0;
								end
							else
								return 39;
							end
						else
							return CalcRatAB(CalcStat("BPEPRatPA",L),CalcStat("BPEPRatPB",L),CalcStat("BPEPRatPCapR",L),N);
						end
					else
						return CalcStat("BPEPRatPB",L);
					end
				elseif SN > "BPEPRATPCAP" then
					if SN < "BRATPARTBPE" then
						if SN < "BRATEXTRA" then
							if SN < "BPET" then
								if SN == "BPEPRATPCAPR" then
									return CalcStat("BPEPRatPB",L)*CalcStat("BPEPRatPC",L);
								else
									return 0;
								end
							elseif SN > "BPET" then
								if SN > "BRATCRITMAGN" then
									if SN == "BRATDEVHIT" then
										if Lm <= 50 then
											return LinFmod(1,400,4000,1,50,L);
										elseif Lm <= 60 then
											return LinFmod(1,4000,6000,50,60,L);
										elseif Lm <= 65 then
											return LinFmod(1,6000,8000,60,65,L);
										elseif Lm <= 75 then
											return LinFmod(1,8000,12000,65,75,L);
										elseif Lm <= 85 then
											return LinFmod(1,12000,18000,75,85,L);
										elseif Lm <= 95 then
											return LinFmod(1,18000,26000,85,95,L);
										elseif Lm <= 100 then
											return LinFmod(1,26000,36000,95,100,L);
										elseif Lm <= 105 then
											return LinFmod(1,36000,48000,100,105,L);
										elseif Lm <= 115 then
											return LinFmod(1,53000,72000,106,115,L);
										elseif Lm <= 120 then
											return LinFmod(1,83000,90000,116,120,L);
										elseif Lm <= 130 then
											return LinFmod(1,104000,135000,121,130,L);
										elseif Lm <= 140 then
											return LinFmod(1,155000,270000,131,140,L);
										else
											return LinFmod(1,311000,540000,141,150,L);
										end
									else
										return 0;
									end
								elseif SN == "BRATCRITMAGN" then
									if Lm <= 50 then
										return LinFmod(1,700,7000,1,50,L);
									elseif Lm <= 60 then
										return LinFmod(1,7000,10500,50,60,L);
									elseif Lm <= 65 then
										return LinFmod(1,10500,14000,60,65,L);
									elseif Lm <= 75 then
										return LinFmod(1,14000,21000,65,75,L);
									elseif Lm <= 85 then
										return LinFmod(1,21000,31500,75,85,L);
									elseif Lm <= 95 then
										return LinFmod(1,31500,46000,85,95,L);
									elseif Lm <= 100 then
										return LinFmod(1,46000,64000,95,100,L);
									elseif Lm <= 105 then
										return LinFmod(1,64000,85000,100,105,L);
									elseif Lm <= 115 then
										return LinFmod(1,94000,128000,106,115,L);
									elseif Lm <= 120 then
										return LinFmod(1,147000,160000,116,120,L);
									elseif Lm <= 130 then
										return LinFmod(1,184000,240000,121,130,L);
									elseif Lm <= 140 then
										return LinFmod(1,276000,480000,131,140,L);
									else
										return LinFmod(1,552000,960000,141,150,L);
									end
								else
									return 0;
								end
							else
								return StatLinInter("PntMPBPE","TraitProg","ProgBBPE","",L,N,0);
							end
						elseif SN > "BRATEXTRA" then
							if SN < "BRATMITLIGHT" then
								if SN == "BRATMITHEAVY" then
									return CalcStat("BRatStandard",L);
								else
									return 0;
								end
							elseif SN > "BRATMITLIGHT" then
								if SN > "BRATMITMEDIUM" then
									if SN == "BRATOUTHEAL" then
										if Lm <= 50 then
											return LinFmod(1,450,4500,1,50,L);
										elseif Lm <= 60 then
											return LinFmod(1,4500,6800,50,60,L);
										elseif Lm <= 65 then
											return LinFmod(1,6800,9100,60,65,L);
										elseif Lm <= 75 then
											return LinFmod(1,9100,13700,65,75,L);
										elseif Lm <= 85 then
											return LinFmod(1,13700,20600,75,85,L);
										elseif Lm <= 95 then
											return LinFmod(1,20600,29800,85,95,L);
										elseif Lm <= 100 then
											return LinFmod(1,29800,41000,95,100,L);
										elseif Lm <= 105 then
											return LinFmod(1,41000,55000,100,105,L);
										elseif Lm <= 115 then
											return LinFmod(1,61000,83000,106,115,L);
										elseif Lm <= 120 then
											return LinFmod(1,95000,104000,116,120,L);
										elseif Lm <= 130 then
											return LinFmod(1,120000,156000,121,130,L);
										elseif Lm <= 140 then
											return LinFmod(1,179000,312000,131,140,L);
										else
											return LinFmod(1,360000,620000,141,150,L);
										end
									else
										return 0;
									end
								elseif SN == "BRATMITMEDIUM" then
									if Lm <= 50 then
										return LinFmod(1,144,1670,1,50,L);
									elseif Lm <= 60 then
										return LinFmod(1,1670,2500,50,60,L);
									elseif Lm <= 65 then
										return LinFmod(1,2500,3300,60,65,L);
									elseif Lm <= 75 then
										return LinFmod(1,3300,5000,65,75,L);
									elseif Lm <= 85 then
										return LinFmod(1,5000,7500,75,85,L);
									elseif Lm <= 95 then
										return LinFmod(1,7500,10800,85,95,L);
									elseif Lm <= 100 then
										return LinFmod(1,10800,15100,95,100,L);
									elseif Lm <= 105 then
										return LinFmod(1,15100,20100,100,105,L);
									elseif Lm <= 115 then
										return LinFmod(1,22100,30000,106,115,L);
									elseif Lm <= 120 then
										return LinFmod(1,34000,37000,116,120,L);
									elseif Lm <= 130 then
										return LinFmod(1,43000,57000,121,130,L);
									elseif Lm <= 140 then
										return LinFmod(1,65000,113000,131,140,L);
									else
										return LinFmod(1,130000,227000,141,150,L);
									end
								else
									return 0;
								end
							else
								if Lm <= 50 then
									return LinFmod(1,105,1330,1,50,L);
								elseif Lm <= 60 then
									return LinFmod(1,1330,2000,50,60,L);
								elseif Lm <= 65 then
									return LinFmod(1,2000,2660,60,65,L);
								elseif Lm <= 75 then
									return LinFmod(1,2660,4000,65,75,L);
								elseif Lm <= 85 then
									return LinFmod(1,4000,6000,75,85,L);
								elseif Lm <= 95 then
									return LinFmod(1,6000,8700,85,95,L);
								elseif Lm <= 100 then
									return LinFmod(1,8700,12100,95,100,L);
								elseif Lm <= 105 then
									return LinFmod(1,12100,16100,100,105,L);
								elseif Lm <= 115 then
									return LinFmod(1,17600,24000,106,115,L);
								elseif Lm <= 120 then
									return LinFmod(1,27300,30000,116,120,L);
								elseif Lm <= 130 then
									return LinFmod(1,35000,45000,121,130,L);
								elseif Lm <= 140 then
									return LinFmod(1,52000,91000,131,140,L);
								else
									return LinFmod(1,104000,181000,141,150,L);
								end
							end
						else
							if Lm <= 50 then
								return LinFmod(1,300,3000,1,50,L);
							elseif Lm <= 60 then
								return LinFmod(1,3000,4500,50,60,L);
							elseif Lm <= 65 then
								return LinFmod(1,4500,6000,60,65,L);
							elseif Lm <= 75 then
								return LinFmod(1,6000,9000,65,75,L);
							elseif Lm <= 85 then
								return LinFmod(1,9000,13500,75,85,L);
							elseif Lm <= 95 then
								return LinFmod(1,13500,19500,85,95,L);
							elseif Lm <= 100 then
								return LinFmod(1,19500,27100,95,100,L);
							elseif Lm <= 105 then
								return LinFmod(1,27100,36000,100,105,L);
							elseif Lm <= 115 then
								return LinFmod(1,40000,54000,106,115,L);
							elseif Lm <= 120 then
								return LinFmod(1,62000,68000,116,120,L);
							elseif Lm <= 130 then
								return LinFmod(1,78000,102000,121,130,L);
							elseif Lm <= 140 then
								return LinFmod(1,117000,204000,131,140,L);
							else
								return LinFmod(1,235000,410000,141,150,L);
							end
						end
					elseif SN > "BRATPARTBPE" then
						if SN < "BRAWLERCDHASPOWER" then
							if SN < "BRATROUNDED" then
								if SN == "BRATPROGB" then
									if Lm <= 50 then
										return RoundDbl(CalcStat(C,L));
									else
										return CalcStat(C,L);
									end
								else
									return 0;
								end
							elseif SN > "BRATROUNDED" then
								if SN > "BRATSTANDARD" then
									if SN == "BRAWLERCDARMOURTYPE" then
										return 3;
									else
										return 0;
									end
								elseif SN == "BRATSTANDARD" then
									if Lm <= 50 then
										return LinFmod(1,200,2000,1,50,L);
									elseif Lm <= 60 then
										return LinFmod(1,2000,3000,50,60,L);
									elseif Lm <= 65 then
										return LinFmod(1,3000,4000,60,65,L);
									elseif Lm <= 75 then
										return LinFmod(1,4000,6000,65,75,L);
									elseif Lm <= 85 then
										return LinFmod(1,6000,9000,75,85,L);
									elseif Lm <= 95 then
										return LinFmod(1,9000,13000,85,95,L);
									elseif Lm <= 100 then
										return LinFmod(1,13000,18100,95,100,L);
									elseif Lm <= 105 then
										return LinFmod(1,18100,24100,100,105,L);
									elseif Lm <= 115 then
										return LinFmod(1,26500,36000,106,115,L);
									elseif Lm <= 120 then
										return LinFmod(1,41000,45000,116,120,L);
									elseif Lm <= 130 then
										return LinFmod(1,52000,68000,121,130,L);
									elseif Lm <= 140 then
										return LinFmod(1,78000,136000,131,140,L);
									else
										return LinFmod(1,156000,272000,141,150,L);
									end
								else
									return 0;
								end
							else
								if Lm <= 50 then
									return RoundDbl(CalcStat(C,L));
								elseif Lm <= 105 then
									return RoundDbl(CalcStat(C,L),-1);
								elseif Lm <= 115 then
									return RoundDbl(CalcStat(C,L),-2);
								elseif Lm <= 130 then
									return RoundDbl(CalcStat(C,L),-1);
								else
									return RoundDbl(CalcStat(C,L),-2);
								end
							end
						elseif SN > "BRAWLERCDHASPOWER" then
							if SN < "CAPTAINCDARMOURTYPE" then
								if SN > "BURGLARCDARMOURTYPE" then
									if SN == "BURGLARCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "BURGLARCDARMOURTYPE" then
									return 2;
								else
									return 0;
								end
							elseif SN > "CAPTAINCDARMOURTYPE" then
								if SN > "CAPTAINCDCANBLOCK" then
									if SN == "CAPTAINCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "CAPTAINCDCANBLOCK" then
									if Lm <= 14 then
										return 0;
									else
										return 1;
									end
								else
									return 0;
								end
							else
								return 3;
							end
						else
							return 1;
						end
					else
						if Lm <= 50 then
							return LinFmod(1,350,3500,1,50,L);
						elseif Lm <= 60 then
							return LinFmod(1,3500,5300,50,60,L);
						elseif Lm <= 65 then
							return LinFmod(1,5300,7100,60,65,L);
						elseif Lm <= 75 then
							return LinFmod(1,7100,10700,65,75,L);
						elseif Lm <= 85 then
							return LinFmod(1,10700,16100,75,85,L);
						elseif Lm <= 95 then
							return LinFmod(1,16100,23300,85,95,L);
						elseif Lm <= 100 then
							return LinFmod(1,23300,32000,95,100,L);
						elseif Lm <= 105 then
							return LinFmod(1,32000,43000,100,105,L);
						elseif Lm <= 115 then
							return LinFmod(1,47000,65000,106,115,L);
						elseif Lm <= 120 then
							return LinFmod(1,75000,81000,116,120,L);
						elseif Lm <= 130 then
							return LinFmod(1,93000,122000,121,130,L);
						elseif Lm <= 140 then
							return LinFmod(1,140000,244000,131,140,L);
						else
							return LinFmod(1,281000,488000,141,150,L);
						end
					end
				else
					return 13;
				end
			elseif SN > "CHAMPIONCDARMOURTYPE" then
				if SN < "CRITMAGNPPRAT" then
					if SN < "CRITDEFPRATPCAPR" then
						if SN < "CRITDEFPRATP" then
							if SN < "CHAMPIONCDHASPOWER" then
								if SN == "CHAMPIONCDCANBLOCK" then
									if Lm <= 9 then
										return 0;
									else
										return 1;
									end
								else
									return 0;
								end
							elseif SN > "CHAMPIONCDHASPOWER" then
								if SN > "CRITDEFPBONUS" then
									if SN == "CRITDEFPPRAT" then
										return CalcRatAB(CalcStat("CritDefPRatPA",L),CalcStat("CritDefPRatPB",L),CalcStat("CritDefPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "CRITDEFPBONUS" then
									return 0;
								else
									return 0;
								end
							else
								return 1;
							end
						elseif SN > "CRITDEFPRATP" then
							if SN < "CRITDEFPRATPB" then
								if SN == "CRITDEFPRATPA" then
									return 240;
								else
									return 0;
								end
							elseif SN > "CRITDEFPRATPB" then
								if SN > "CRITDEFPRATPC" then
									if SN == "CRITDEFPRATPCAP" then
										return 80;
									else
										return 0;
									end
								elseif SN == "CRITDEFPRATPC" then
									return 0.5;
								else
									return 0;
								end
							else
								return CalcStat("BRatRounded",L,"BRatStandard");
							end
						else
							return CalcPercAB(CalcStat("CritDefPRatPA",L),CalcStat("CritDefPRatPB",L),CalcStat("CritDefPRatPCap",L),N);
						end
					elseif SN > "CRITDEFPRATPCAPR" then
						if SN < "CRITHITPRATPB" then
							if SN < "CRITHITPPRAT" then
								if SN == "CRITHITPBONUS" then
									return 0;
								else
									return 0;
								end
							elseif SN > "CRITHITPPRAT" then
								if SN > "CRITHITPRATP" then
									if SN == "CRITHITPRATPA" then
										return 75;
									else
										return 0;
									end
								elseif SN == "CRITHITPRATP" then
									return CalcPercAB(CalcStat("CritHitPRatPA",L),CalcStat("CritHitPRatPB",L),CalcStat("CritHitPRatPCap",L),N);
								else
									return 0;
								end
							else
								return CalcRatAB(CalcStat("CritHitPRatPA",L),CalcStat("CritHitPRatPB",L),CalcStat("CritHitPRatPCapR",L),N);
							end
						elseif SN > "CRITHITPRATPB" then
							if SN < "CRITHITPRATPCAP" then
								if SN == "CRITHITPRATPC" then
									return 0.5;
								else
									return 0;
								end
							elseif SN > "CRITHITPRATPCAP" then
								if SN > "CRITHITPRATPCAPR" then
									if SN == "CRITMAGNPBONUS" then
										return 0;
									else
										return 0;
									end
								elseif SN == "CRITHITPRATPCAPR" then
									return CalcStat("CritHitPRatPB",L)*CalcStat("CritHitPRatPC",L);
								else
									return 0;
								end
							else
								return 25;
							end
						else
							return CalcStat("BRatRounded",L,"BRatStandard");
						end
					else
						return CalcStat("CritDefPRatPB",L)*CalcStat("CritDefPRatPC",L);
					end
				elseif SN > "CRITMAGNPPRAT" then
					if SN < "DEVHITPRATPA" then
						if SN < "CRITMAGNPRATPCAP" then
							if SN < "CRITMAGNPRATPA" then
								if SN == "CRITMAGNPRATP" then
									return CalcPercAB(CalcStat("CritMagnPRatPA",L),CalcStat("CritMagnPRatPB",L),CalcStat("CritMagnPRatPCap",L),N);
								else
									return 0;
								end
							elseif SN > "CRITMAGNPRATPA" then
								if SN > "CRITMAGNPRATPB" then
									if SN == "CRITMAGNPRATPC" then
										return 0.5;
									else
										return 0;
									end
								elseif SN == "CRITMAGNPRATPB" then
									return CalcStat("BRatRounded",L,"BRatCritMagn");
								else
									return 0;
								end
							else
								return 225;
							end
						elseif SN > "CRITMAGNPRATPCAP" then
							if SN < "DEVHITPBONUS" then
								if SN == "CRITMAGNPRATPCAPR" then
									return CalcStat("CritMagnPRatPB",L)*CalcStat("CritMagnPRatPC",L);
								else
									return 0;
								end
							elseif SN > "DEVHITPBONUS" then
								if SN > "DEVHITPPRAT" then
									if SN == "DEVHITPRATP" then
										return CalcPercAB(CalcStat("DevHitPRatPA",L),CalcStat("DevHitPRatPB",L),CalcStat("DevHitPRatPCap",L),N);
									else
										return 0;
									end
								elseif SN == "DEVHITPPRAT" then
									return CalcRatAB(CalcStat("DevHitPRatPA",L),CalcStat("DevHitPRatPB",L),CalcStat("DevHitPRatPCapR",L),N);
								else
									return 0;
								end
							else
								return 0;
							end
						else
							return 75;
						end
					elseif SN > "DEVHITPRATPA" then
						if SN < "EVADEPBONUS" then
							if SN < "DEVHITPRATPC" then
								if SN == "DEVHITPRATPB" then
									return CalcStat("BRatRounded",L,"BRatDevHit");
								else
									return 0;
								end
							elseif SN > "DEVHITPRATPC" then
								if SN > "DEVHITPRATPCAP" then
									if SN == "DEVHITPRATPCAPR" then
										return CalcStat("DevHitPRatPB",L)*CalcStat("DevHitPRatPC",L);
									else
										return 0;
									end
								elseif SN == "DEVHITPRATPCAP" then
									return 10;
								else
									return 0;
								end
							else
								return 0.5;
							end
						elseif SN > "EVADEPBONUS" then
							if SN < "EVADEPRATPA" then
								if SN > "EVADEPPRAT" then
									if SN == "EVADEPRATP" then
										return CalcStat("BPEPRatP",L,N);
									else
										return 0;
									end
								elseif SN == "EVADEPPRAT" then
									return CalcStat("BPEPPRat",L,N);
								else
									return 0;
								end
							elseif SN > "EVADEPRATPA" then
								if SN > "EVADEPRATPB" then
									if SN == "EVADEPRATPC" then
										return CalcStat("BPEPRatPC",L);
									else
										return 0;
									end
								elseif SN == "EVADEPRATPB" then
									return CalcStat("BPEPRatPB",L);
								else
									return 0;
								end
							else
								return CalcStat("BPEPRatPA",L);
							end
						else
							return CalcStat("BPEPBonus",L);
						end
					else
						return 30;
					end
				else
					return CalcRatAB(CalcStat("CritMagnPRatPA",L),CalcStat("CritMagnPRatPB",L),CalcStat("CritMagnPRatPCapR",L),N);
				end
			else
				return 3;
			end
		elseif SN > "EVADEPRATPCAP" then
			if SN < "MITHEAVYPBONUS" then
				if SN < "INDMGPRATPC" then
					if SN < "GUARDIANCDARMOURTYPE" then
						if SN < "FINESSEPRATPA" then
							if SN < "FINESSEPBONUS" then
								if SN == "EVADEPRATPCAPR" then
									return CalcStat("BPEPRatPCapR",L);
								else
									return 0;
								end
							elseif SN > "FINESSEPBONUS" then
								if SN > "FINESSEPPRAT" then
									if SN == "FINESSEPRATP" then
										return CalcPercAB(CalcStat("FinessePRatPA",L),CalcStat("FinessePRatPB",L),CalcStat("FinessePRatPCap",L),N);
									else
										return 0;
									end
								elseif SN == "FINESSEPPRAT" then
									return CalcRatAB(CalcStat("FinessePRatPA",L),CalcStat("FinessePRatPB",L),CalcStat("FinessePRatPCapR",L),N);
								else
									return 0;
								end
							else
								return 0;
							end
						elseif SN > "FINESSEPRATPA" then
							if SN < "FINESSEPRATPC" then
								if SN == "FINESSEPRATPB" then
									return CalcStat("BRatRounded",L,"BRatStandard");
								else
									return 0;
								end
							elseif SN > "FINESSEPRATPC" then
								if SN > "FINESSEPRATPCAP" then
									if SN == "FINESSEPRATPCAPR" then
										return CalcStat("FinessePRatPB",L)*CalcStat("FinessePRatPC",L);
									else
										return 0;
									end
								elseif SN == "FINESSEPRATPCAP" then
									return 50;
								else
									return 0;
								end
							else
								return 0.5;
							end
						else
							return 150;
						end
					elseif SN > "GUARDIANCDARMOURTYPE" then
						if SN < "INDMGPBONUS" then
							if SN < "GUARDIANCDHASPOWER" then
								if SN == "GUARDIANCDCANBLOCK" then
									return 1;
								else
									return 0;
								end
							elseif SN > "GUARDIANCDHASPOWER" then
								if SN > "HUNTERCDARMOURTYPE" then
									if SN == "HUNTERCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "HUNTERCDARMOURTYPE" then
									return 2;
								else
									return 0;
								end
							else
								return 1;
							end
						elseif SN > "INDMGPBONUS" then
							if SN < "INDMGPRATP" then
								if SN == "INDMGPPRAT" then
									return CalcRatAB(CalcStat("InDmgPRatPA",L),CalcStat("InDmgPRatPB",L),CalcStat("InDmgPRatPCapR",L),N);
								else
									return 0;
								end
							elseif SN > "INDMGPRATP" then
								if SN > "INDMGPRATPA" then
									if SN == "INDMGPRATPB" then
										return CalcStat("BRatRounded",L,"BRatStandard");
									else
										return 0;
									end
								elseif SN == "INDMGPRATPA" then
									return 1200;
								else
									return 0;
								end
							else
								return CalcPercAB(CalcStat("InDmgPRatPA",L),CalcStat("InDmgPRatPB",L),CalcStat("InDmgPRatPCap",L),N);
							end
						else
							return 0;
						end
					else
						return 3;
					end
				elseif SN > "INDMGPRATPC" then
					if SN < "INHEALPRATPCAPR" then
						if SN < "INHEALPRATP" then
							if SN < "INDMGPRATPCAPR" then
								if SN == "INDMGPRATPCAP" then
									return 400;
								else
									return 0;
								end
							elseif SN > "INDMGPRATPCAPR" then
								if SN > "INHEALPBONUS" then
									if SN == "INHEALPPRAT" then
										return CalcRatAB(CalcStat("InHealPRatPA",L),CalcStat("InHealPRatPB",L),CalcStat("InHealPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "INHEALPBONUS" then
									return 0;
								else
									return 0;
								end
							else
								return CalcStat("InDmgPRatPB",L)*CalcStat("InDmgPRatPC",L);
							end
						elseif SN > "INHEALPRATP" then
							if SN < "INHEALPRATPB" then
								if SN == "INHEALPRATPA" then
									return 75;
								else
									return 0;
								end
							elseif SN > "INHEALPRATPB" then
								if SN > "INHEALPRATPC" then
									if SN == "INHEALPRATPCAP" then
										return 25;
									else
										return 0;
									end
								elseif SN == "INHEALPRATPC" then
									return 0.5;
								else
									return 0;
								end
							else
								return CalcStat("BRatRounded",L,"BRatStandard");
							end
						else
							return CalcPercAB(CalcStat("InHealPRatPA",L),CalcStat("InHealPRatPB",L),CalcStat("InHealPRatPCap",L),N);
						end
					elseif SN > "INHEALPRATPCAPR" then
						if SN < "LVLEXPCOSTTOT" then
							if SN < "LOREMASTERCDARMOURTYPE" then
								if SN == "LEVELCAP" then
									return 140;
								else
									return 0;
								end
							elseif SN > "LOREMASTERCDARMOURTYPE" then
								if SN > "LOREMASTERCDHASPOWER" then
									if SN == "LVLEXPCOST" then
										if Lm <= 1 then
											return 0;
										elseif Lm <= 5 then
											return RoundDbl(12.5*L*L+12.5666666666667*L+24.8666666666667);
										elseif Lm <= 10 then
											return RoundDbl(33.8*L*L-179.48*L+452.6);
										elseif Lm <= 15 then
											return RoundDbl(55.05*L*L-583.77*L+2370.5);
										elseif Lm <= 20 then
											return RoundDbl(76.2*L*L-1196.96*L+6809);
										elseif Lm <= 25 then
											return RoundDbl(97.4*L*L-2023*L+14849.8);
										elseif Lm <= 30 then
											return RoundDbl(118.7*L*L-3066.02 *L+27612.8);
										elseif Lm <= 35 then
											return RoundDbl(139.95*L*L-4319.23*L+46084.1);
										elseif Lm <= 40 then
											return RoundDbl(161.2*L*L-5785.04*L+71356.2);
										elseif Lm <= 45 then
											return RoundDbl(182.5*L*L-7467.38*L+104569.8);
										elseif Lm <= 50 then
											return RoundDbl(203.8*L*L-9363.48*L+146761.8);
										elseif Lm <= 55 then
											return RoundDbl(225.05*L*L-11467.77*L+198851.3);
										elseif Lm <= 60 then
											return RoundDbl(246.3*L*L-13784.46*L+261988);
										elseif Lm <= 70 then
											return RoundDbl(ExpFmod(CalcStat("LvlExpCost",60),61,5.071,L,nil,3.485));
										elseif Lm <= 75 then
											return RoundDbl(ExpFmod(CalcStat("LvlExpCost",70),71,5.072,L,nil,-0.95));
										else
											return ExpFmod(CalcStat("LvlExpCost",75),76,5,L,0,-0.5);
										end
									else
										return 0;
									end
								elseif SN == "LOREMASTERCDHASPOWER" then
									return 1;
								else
									return 0;
								end
							else
								return 1;
							end
						elseif SN > "LVLEXPCOSTTOT" then
							if SN < "MINSTRELCDARMOURTYPE" then
								if SN > "MARINERCDARMOURTYPE" then
									if SN == "MARINERCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "MARINERCDARMOURTYPE" then
									return 2;
								else
									return 0;
								end
							elseif SN > "MINSTRELCDARMOURTYPE" then
								if SN > "MINSTRELCDCANBLOCK" then
									if SN == "MINSTRELCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "MINSTRELCDCANBLOCK" then
									if Lm <= 19 then
										return 0;
									else
										return 1;
									end
								else
									return 0;
								end
							else
								return 1;
							end
						else
							if Lm <= 1 then
								return 0;
							else
								return CalcStat("LvlExpCostTot",L-1)+CalcStat("LvlExpCost",L);
							end
						end
					else
						return CalcStat("InHealPRatPB",L)*CalcStat("InHealPRatPC",L);
					end
				else
					return 0.5;
				end
			elseif SN > "MITHEAVYPBONUS" then
				if SN < "MITMEDIUMPRATPC" then
					if SN < "MITLIGHTPRATP" then
						if SN < "MITHEAVYPRATPC" then
							if SN < "MITHEAVYPRATP" then
								if SN == "MITHEAVYPPRAT" then
									return CalcRatAB(CalcStat("MitHeavyPRatPA",L),CalcStat("MitHeavyPRatPB",L),CalcStat("MitHeavyPRatPCapR",L),N);
								else
									return 0;
								end
							elseif SN > "MITHEAVYPRATP" then
								if SN > "MITHEAVYPRATPA" then
									if SN == "MITHEAVYPRATPB" then
										return CalcStat("BRatRounded",L,"BRatMitHeavy");
									else
										return 0;
									end
								elseif SN == "MITHEAVYPRATPA" then
									return 180;
								else
									return 0;
								end
							else
								return CalcPercAB(CalcStat("MitHeavyPRatPA",L),CalcStat("MitHeavyPRatPB",L),CalcStat("MitHeavyPRatPCap",L),N);
							end
						elseif SN > "MITHEAVYPRATPC" then
							if SN < "MITHEAVYPRATPCAPR" then
								if SN == "MITHEAVYPRATPCAP" then
									return 60;
								else
									return 0;
								end
							elseif SN > "MITHEAVYPRATPCAPR" then
								if SN > "MITLIGHTPBONUS" then
									if SN == "MITLIGHTPPRAT" then
										return CalcRatAB(CalcStat("MitLightPRatPA",L),CalcStat("MitLightPRatPB",L),CalcStat("MitLightPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "MITLIGHTPBONUS" then
									return 0;
								else
									return 0;
								end
							else
								return CalcStat("MitHeavyPRatPB",L)*CalcStat("MitHeavyPRatPC",L);
							end
						else
							return 0.5;
						end
					elseif SN > "MITLIGHTPRATP" then
						if SN < "MITLIGHTPRATPCAPR" then
							if SN < "MITLIGHTPRATPB" then
								if SN == "MITLIGHTPRATPA" then
									return 120;
								else
									return 0;
								end
							elseif SN > "MITLIGHTPRATPB" then
								if SN > "MITLIGHTPRATPC" then
									if SN == "MITLIGHTPRATPCAP" then
										return 40;
									else
										return 0;
									end
								elseif SN == "MITLIGHTPRATPC" then
									return 0.5;
								else
									return 0;
								end
							else
								return CalcStat("BRatRounded",L,"BRatMitLight");
							end
						elseif SN > "MITLIGHTPRATPCAPR" then
							if SN < "MITMEDIUMPRATP" then
								if SN > "MITMEDIUMPBONUS" then
									if SN == "MITMEDIUMPPRAT" then
										return CalcRatAB(CalcStat("MitMediumPRatPA",L),CalcStat("MitMediumPRatPB",L),CalcStat("MitMediumPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "MITMEDIUMPBONUS" then
									return 0;
								else
									return 0;
								end
							elseif SN > "MITMEDIUMPRATP" then
								if SN > "MITMEDIUMPRATPA" then
									if SN == "MITMEDIUMPRATPB" then
										return CalcStat("BRatRounded",L,"BRatMitMedium");
									else
										return 0;
									end
								elseif SN == "MITMEDIUMPRATPA" then
									return 150;
								else
									return 0;
								end
							else
								return CalcPercAB(CalcStat("MitMediumPRatPA",L),CalcStat("MitMediumPRatPB",L),CalcStat("MitMediumPRatPCap",L),N);
							end
						else
							return CalcStat("MitLightPRatPB",L)*CalcStat("MitLightPRatPC",L);
						end
					else
						return CalcPercAB(CalcStat("MitLightPRatPA",L),CalcStat("MitLightPRatPB",L),CalcStat("MitLightPRatPCap",L),N);
					end
				elseif SN > "MITMEDIUMPRATPC" then
					if SN < "OUTDMGPRATPCAPR" then
						if SN < "OUTDMGPRATP" then
							if SN < "MITMEDIUMPRATPCAPR" then
								if SN == "MITMEDIUMPRATPCAP" then
									return 50;
								else
									return 0;
								end
							elseif SN > "MITMEDIUMPRATPCAPR" then
								if SN > "OUTDMGPBONUS" then
									if SN == "OUTDMGPPRAT" then
										return CalcRatAB(CalcStat("OutDmgPRatPA",L),CalcStat("OutDmgPRatPB",L),CalcStat("OutDmgPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "OUTDMGPBONUS" then
									return 0;
								else
									return 0;
								end
							else
								return CalcStat("MitMediumPRatPB",L)*CalcStat("MitMediumPRatPC",L);
							end
						elseif SN > "OUTDMGPRATP" then
							if SN < "OUTDMGPRATPB" then
								if SN == "OUTDMGPRATPA" then
									return 600;
								else
									return 0;
								end
							elseif SN > "OUTDMGPRATPB" then
								if SN > "OUTDMGPRATPC" then
									if SN == "OUTDMGPRATPCAP" then
										return 200;
									else
										return 0;
									end
								elseif SN == "OUTDMGPRATPC" then
									return 0.5;
								else
									return 0;
								end
							else
								return CalcStat("BRatRounded",L,"BRatExtra");
							end
						else
							return CalcPercAB(CalcStat("OutDmgPRatPA",L),CalcStat("OutDmgPRatPB",L),CalcStat("OutDmgPRatPCap",L),N);
						end
					elseif SN > "OUTDMGPRATPCAPR" then
						if SN < "OUTHEALPRATPB" then
							if SN < "OUTHEALPPRAT" then
								if SN == "OUTHEALPBONUS" then
									return 0;
								else
									return 0;
								end
							elseif SN > "OUTHEALPPRAT" then
								if SN > "OUTHEALPRATP" then
									if SN == "OUTHEALPRATPA" then
										return 210;
									else
										return 0;
									end
								elseif SN == "OUTHEALPRATP" then
									return CalcPercAB(CalcStat("OutHealPRatPA",L),CalcStat("OutHealPRatPB",L),CalcStat("OutHealPRatPCap",L),N);
								else
									return 0;
								end
							else
								return CalcRatAB(CalcStat("OutHealPRatPA",L),CalcStat("OutHealPRatPB",L),CalcStat("OutHealPRatPCapR",L),N);
							end
						elseif SN > "OUTHEALPRATPB" then
							if SN < "OUTHEALPRATPCAPR" then
								if SN > "OUTHEALPRATPC" then
									if SN == "OUTHEALPRATPCAP" then
										return 70;
									else
										return 0;
									end
								elseif SN == "OUTHEALPRATPC" then
									return 0.5;
								else
									return 0;
								end
							elseif SN > "OUTHEALPRATPCAPR" then
								if SN > "PARRYPBONUS" then
									if SN == "PARRYPPRAT" then
										return CalcStat("BPEPPRat",L,N);
									else
										return 0;
									end
								elseif SN == "PARRYPBONUS" then
									return CalcStat("BPEPBonus",L);
								else
									return 0;
								end
							else
								return CalcStat("OutHealPRatPB",L)*CalcStat("OutHealPRatPC",L);
							end
						else
							return CalcStat("BRatRounded",L,"BRatOutHeal");
						end
					else
						return CalcStat("OutDmgPRatPB",L)*CalcStat("OutDmgPRatPC",L);
					end
				else
					return 0.5;
				end
			else
				return 0;
			end
		else
			return CalcStat("BPEPRatPCap",L);
		end
	elseif SN > "PARRYPRATP" then
		if SN < "PHYMITHPRATPB" then
			if SN < "PARTEVADEPRATPA" then
				if SN < "PARTBLOCKPRATPCAP" then
					if SN < "PARTBLOCKMITPRATPB" then
						if SN < "PARRYPRATPCAPR" then
							if SN < "PARRYPRATPB" then
								if SN == "PARRYPRATPA" then
									return CalcStat("BPEPRatPA",L);
								else
									return 0;
								end
							elseif SN > "PARRYPRATPB" then
								if SN > "PARRYPRATPC" then
									if SN == "PARRYPRATPCAP" then
										return CalcStat("BPEPRatPCap",L);
									else
										return 0;
									end
								elseif SN == "PARRYPRATPC" then
									return CalcStat("BPEPRatPC",L);
								else
									return 0;
								end
							else
								return CalcStat("BPEPRatPB",L);
							end
						elseif SN > "PARRYPRATPCAPR" then
							if SN < "PARTBLOCKMITPPRAT" then
								if SN == "PARTBLOCKMITPBONUS" then
									return CalcStat("PartMitPBonus",L);
								else
									return 0;
								end
							elseif SN > "PARTBLOCKMITPPRAT" then
								if SN > "PARTBLOCKMITPRATP" then
									if SN == "PARTBLOCKMITPRATPA" then
										return CalcStat("PartMitPRatPA",L);
									else
										return 0;
									end
								elseif SN == "PARTBLOCKMITPRATP" then
									return CalcStat("PartMitPRatP",L,N);
								else
									return 0;
								end
							else
								return CalcStat("PartMitPPRat",L,N);
							end
						else
							return CalcStat("BPEPRatPCapR",L);
						end
					elseif SN > "PARTBLOCKMITPRATPB" then
						if SN < "PARTBLOCKPPRAT" then
							if SN < "PARTBLOCKMITPRATPCAP" then
								if SN == "PARTBLOCKMITPRATPC" then
									return CalcStat("PartMitPRatPC",L);
								else
									return 0;
								end
							elseif SN > "PARTBLOCKMITPRATPCAP" then
								if SN > "PARTBLOCKMITPRATPCAPR" then
									if SN == "PARTBLOCKPBONUS" then
										return CalcStat("PartBPEPBonus",L);
									else
										return 0;
									end
								elseif SN == "PARTBLOCKMITPRATPCAPR" then
									return CalcStat("PartMitPRatPCapR",L);
								else
									return 0;
								end
							else
								return CalcStat("PartMitPRatPCap",L);
							end
						elseif SN > "PARTBLOCKPPRAT" then
							if SN < "PARTBLOCKPRATPA" then
								if SN == "PARTBLOCKPRATP" then
									return CalcStat("PartBPEPRatP",L,N);
								else
									return 0;
								end
							elseif SN > "PARTBLOCKPRATPA" then
								if SN > "PARTBLOCKPRATPB" then
									if SN == "PARTBLOCKPRATPC" then
										return CalcStat("PartBPEPRatPC",L);
									else
										return 0;
									end
								elseif SN == "PARTBLOCKPRATPB" then
									return CalcStat("PartBPEPRatPB",L);
								else
									return 0;
								end
							else
								return CalcStat("PartBPEPRatPA",L);
							end
						else
							return CalcStat("PartBPEPPRat",L,N);
						end
					else
						return CalcStat("PartMitPRatPB",L);
					end
				elseif SN > "PARTBLOCKPRATPCAP" then
					if SN < "PARTEVADEMITPBONUS" then
						if SN < "PARTBPEPRATPA" then
							if SN < "PARTBPEPBONUS" then
								if SN == "PARTBLOCKPRATPCAPR" then
									return CalcStat("PartBPEPRatPCapR",L);
								else
									return 0;
								end
							elseif SN > "PARTBPEPBONUS" then
								if SN > "PARTBPEPPRAT" then
									if SN == "PARTBPEPRATP" then
										return CalcPercAB(CalcStat("PartBPEPRatPA",L),CalcStat("PartBPEPRatPB",L),CalcStat("PartBPEPRatPCap",L),N);
									else
										return 0;
									end
								elseif SN == "PARTBPEPPRAT" then
									return CalcRatAB(CalcStat("PartBPEPRatPA",L),CalcStat("PartBPEPRatPB",L),CalcStat("PartBPEPRatPCapR",L),N);
								else
									return 0;
								end
							else
								return 0;
							end
						elseif SN > "PARTBPEPRATPA" then
							if SN < "PARTBPEPRATPC" then
								if SN == "PARTBPEPRATPB" then
									return CalcStat("BRatRounded",L,"BRatPartBPE");
								else
									return 0;
								end
							elseif SN > "PARTBPEPRATPC" then
								if SN > "PARTBPEPRATPCAP" then
									if SN == "PARTBPEPRATPCAPR" then
										return CalcStat("PartBPEPRatPB",L)*CalcStat("PartBPEPRatPC",L);
									else
										return 0;
									end
								elseif SN == "PARTBPEPRATPCAP" then
									return 25;
								else
									return 0;
								end
							else
								return 0.5;
							end
						else
							return 75;
						end
					elseif SN > "PARTEVADEMITPBONUS" then
						if SN < "PARTEVADEMITPRATPC" then
							if SN < "PARTEVADEMITPRATP" then
								if SN == "PARTEVADEMITPPRAT" then
									return CalcStat("PartMitPPRat",L,N);
								else
									return 0;
								end
							elseif SN > "PARTEVADEMITPRATP" then
								if SN > "PARTEVADEMITPRATPA" then
									if SN == "PARTEVADEMITPRATPB" then
										return CalcStat("PartMitPRatPB",L);
									else
										return 0;
									end
								elseif SN == "PARTEVADEMITPRATPA" then
									return CalcStat("PartMitPRatPA",L);
								else
									return 0;
								end
							else
								return CalcStat("PartMitPRatP",L,N);
							end
						elseif SN > "PARTEVADEMITPRATPC" then
							if SN < "PARTEVADEPBONUS" then
								if SN > "PARTEVADEMITPRATPCAP" then
									if SN == "PARTEVADEMITPRATPCAPR" then
										return CalcStat("PartMitPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "PARTEVADEMITPRATPCAP" then
									return CalcStat("PartMitPRatPCap",L);
								else
									return 0;
								end
							elseif SN > "PARTEVADEPBONUS" then
								if SN > "PARTEVADEPPRAT" then
									if SN == "PARTEVADEPRATP" then
										return CalcStat("PartBPEPRatP",L,N);
									else
										return 0;
									end
								elseif SN == "PARTEVADEPPRAT" then
									return CalcStat("PartBPEPPRat",L,N);
								else
									return 0;
								end
							else
								return CalcStat("PartBPEPBonus",L);
							end
						else
							return CalcStat("PartMitPRatPC",L);
						end
					else
						return CalcStat("PartMitPBonus",L);
					end
				else
					return CalcStat("PartBPEPRatPCap",L);
				end
			elseif SN > "PARTEVADEPRATPA" then
				if SN < "PARTPARRYMITPRATPCAPR" then
					if SN < "PARTMITPRATPC" then
						if SN < "PARTMITPBONUS" then
							if SN < "PARTEVADEPRATPC" then
								if SN == "PARTEVADEPRATPB" then
									return CalcStat("PartBPEPRatPB",L);
								else
									return 0;
								end
							elseif SN > "PARTEVADEPRATPC" then
								if SN > "PARTEVADEPRATPCAP" then
									if SN == "PARTEVADEPRATPCAPR" then
										return CalcStat("PartBPEPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "PARTEVADEPRATPCAP" then
									return CalcStat("PartBPEPRatPCap",L);
								else
									return 0;
								end
							else
								return CalcStat("PartBPEPRatPC",L);
							end
						elseif SN > "PARTMITPBONUS" then
							if SN < "PARTMITPRATP" then
								if SN == "PARTMITPPRAT" then
									return CalcRatAB(CalcStat("PartMitPRatPA",L),CalcStat("PartMitPRatPB",L),CalcStat("PartMitPRatPCapR",L),N);
								else
									return 0;
								end
							elseif SN > "PARTMITPRATP" then
								if SN > "PARTMITPRATPA" then
									if SN == "PARTMITPRATPB" then
										return CalcStat("BRatRounded",L,"BRatPartBPE");
									else
										return 0;
									end
								elseif SN == "PARTMITPRATPA" then
									return 105;
								else
									return 0;
								end
							else
								return CalcPercAB(CalcStat("PartMitPRatPA",L),CalcStat("PartMitPRatPB",L),CalcStat("PartMitPRatPCap",L),N);
							end
						else
							return 0.1;
						end
					elseif SN > "PARTMITPRATPC" then
						if SN < "PARTPARRYMITPRATP" then
							if SN < "PARTMITPRATPCAPR" then
								if SN == "PARTMITPRATPCAP" then
									return 35;
								else
									return 0;
								end
							elseif SN > "PARTMITPRATPCAPR" then
								if SN > "PARTPARRYMITPBONUS" then
									if SN == "PARTPARRYMITPPRAT" then
										return CalcStat("PartMitPPRat",L,N);
									else
										return 0;
									end
								elseif SN == "PARTPARRYMITPBONUS" then
									return CalcStat("PartMitPBonus",L);
								else
									return 0;
								end
							else
								return CalcStat("PartMitPRatPB",L)*CalcStat("PartMitPRatPC",L);
							end
						elseif SN > "PARTPARRYMITPRATP" then
							if SN < "PARTPARRYMITPRATPB" then
								if SN == "PARTPARRYMITPRATPA" then
									return CalcStat("PartMitPRatPA",L);
								else
									return 0;
								end
							elseif SN > "PARTPARRYMITPRATPB" then
								if SN > "PARTPARRYMITPRATPC" then
									if SN == "PARTPARRYMITPRATPCAP" then
										return CalcStat("PartMitPRatPCap",L);
									else
										return 0;
									end
								elseif SN == "PARTPARRYMITPRATPC" then
									return CalcStat("PartMitPRatPC",L);
								else
									return 0;
								end
							else
								return CalcStat("PartMitPRatPB",L);
							end
						else
							return CalcStat("PartMitPRatP",L,N);
						end
					else
						return 0.5;
					end
				elseif SN > "PARTPARRYMITPRATPCAPR" then
					if SN < "PHYDMGPPRAT" then
						if SN < "PARTPARRYPRATPB" then
							if SN < "PARTPARRYPPRAT" then
								if SN == "PARTPARRYPBONUS" then
									return CalcStat("PartBPEPBonus",L);
								else
									return 0;
								end
							elseif SN > "PARTPARRYPPRAT" then
								if SN > "PARTPARRYPRATP" then
									if SN == "PARTPARRYPRATPA" then
										return CalcStat("PartBPEPRatPA",L);
									else
										return 0;
									end
								elseif SN == "PARTPARRYPRATP" then
									return CalcStat("PartBPEPRatP",L,N);
								else
									return 0;
								end
							else
								return CalcStat("PartBPEPPRat",L,N);
							end
						elseif SN > "PARTPARRYPRATPB" then
							if SN < "PARTPARRYPRATPCAP" then
								if SN == "PARTPARRYPRATPC" then
									return CalcStat("PartBPEPRatPC",L);
								else
									return 0;
								end
							elseif SN > "PARTPARRYPRATPCAP" then
								if SN > "PARTPARRYPRATPCAPR" then
									if SN == "PHYDMGPBONUS" then
										return CalcStat("OutDmgPBonus",L);
									else
										return 0;
									end
								elseif SN == "PARTPARRYPRATPCAPR" then
									return CalcStat("PartBPEPRatPCapR",L);
								else
									return 0;
								end
							else
								return CalcStat("PartBPEPRatPCap",L);
							end
						else
							return CalcStat("PartBPEPRatPB",L);
						end
					elseif SN > "PHYDMGPPRAT" then
						if SN < "PHYDMGPRATPCAP" then
							if SN < "PHYDMGPRATPA" then
								if SN == "PHYDMGPRATP" then
									return CalcStat("OutDmgPRatP",L,N);
								else
									return 0;
								end
							elseif SN > "PHYDMGPRATPA" then
								if SN > "PHYDMGPRATPB" then
									if SN == "PHYDMGPRATPC" then
										return CalcStat("OutDmgPRatPC",L);
									else
										return 0;
									end
								elseif SN == "PHYDMGPRATPB" then
									return CalcStat("OutDmgPRatPB",L);
								else
									return 0;
								end
							else
								return CalcStat("OutDmgPRatPA",L);
							end
						elseif SN > "PHYDMGPRATPCAP" then
							if SN < "PHYMITHPPRAT" then
								if SN > "PHYDMGPRATPCAPR" then
									if SN == "PHYMITHPBONUS" then
										return CalcStat("MitHeavyPBonus",L);
									else
										return 0;
									end
								elseif SN == "PHYDMGPRATPCAPR" then
									return CalcStat("OutDmgPRatPCapR",L);
								else
									return 0;
								end
							elseif SN > "PHYMITHPPRAT" then
								if SN > "PHYMITHPRATP" then
									if SN == "PHYMITHPRATPA" then
										return CalcStat("MitHeavyPRatPA",L);
									else
										return 0;
									end
								elseif SN == "PHYMITHPRATP" then
									return CalcStat("MitHeavyPRatP",L,N);
								else
									return 0;
								end
							else
								return CalcStat("MitHeavyPPRat",L,N);
							end
						else
							return CalcStat("OutDmgPRatPCap",L);
						end
					else
						return CalcStat("OutDmgPPRat",L,N);
					end
				else
					return CalcStat("PartMitPRatPCapR",L);
				end
			else
				return CalcStat("PartBPEPRatPA",L);
			end
		elseif SN > "PHYMITHPRATPB" then
			if SN < "T2PENMIT" then
				if SN < "PNTMPARMOURPENT" then
					if SN < "PHYMITLPRATPCAP" then
						if SN < "PHYMITLPPRAT" then
							if SN < "PHYMITHPRATPCAP" then
								if SN == "PHYMITHPRATPC" then
									return CalcStat("MitHeavyPRatPC",L);
								else
									return 0;
								end
							elseif SN > "PHYMITHPRATPCAP" then
								if SN > "PHYMITHPRATPCAPR" then
									if SN == "PHYMITLPBONUS" then
										return CalcStat("MitLightPBonus",L);
									else
										return 0;
									end
								elseif SN == "PHYMITHPRATPCAPR" then
									return CalcStat("MitHeavyPRatPCapR",L);
								else
									return 0;
								end
							else
								return CalcStat("MitHeavyPRatPCap",L);
							end
						elseif SN > "PHYMITLPPRAT" then
							if SN < "PHYMITLPRATPA" then
								if SN == "PHYMITLPRATP" then
									return CalcStat("MitLightPRatP",L,N);
								else
									return 0;
								end
							elseif SN > "PHYMITLPRATPA" then
								if SN > "PHYMITLPRATPB" then
									if SN == "PHYMITLPRATPC" then
										return CalcStat("MitLightPRatPC",L);
									else
										return 0;
									end
								elseif SN == "PHYMITLPRATPB" then
									return CalcStat("MitLightPRatPB",L);
								else
									return 0;
								end
							else
								return CalcStat("MitLightPRatPA",L);
							end
						else
							return CalcStat("MitLightPPRat",L,N);
						end
					elseif SN > "PHYMITLPRATPCAP" then
						if SN < "PHYMITMPRATPA" then
							if SN < "PHYMITMPBONUS" then
								if SN == "PHYMITLPRATPCAPR" then
									return CalcStat("MitLightPRatPCapR",L);
								else
									return 0;
								end
							elseif SN > "PHYMITMPBONUS" then
								if SN > "PHYMITMPPRAT" then
									if SN == "PHYMITMPRATP" then
										return CalcStat("MitMediumPRatP",L,N);
									else
										return 0;
									end
								elseif SN == "PHYMITMPPRAT" then
									return CalcStat("MitMediumPPRat",L,N);
								else
									return 0;
								end
							else
								return CalcStat("MitMediumPBonus",L);
							end
						elseif SN > "PHYMITMPRATPA" then
							if SN < "PHYMITMPRATPC" then
								if SN == "PHYMITMPRATPB" then
									return CalcStat("MitMediumPRatPB",L);
								else
									return 0;
								end
							elseif SN > "PHYMITMPRATPC" then
								if SN > "PHYMITMPRATPCAP" then
									if SN == "PHYMITMPRATPCAPR" then
										return CalcStat("MitMediumPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "PHYMITMPRATPCAP" then
									return CalcStat("MitMediumPRatPCap",L);
								else
									return 0;
								end
							else
								return CalcStat("MitMediumPRatPC",L);
							end
						else
							return CalcStat("MitMediumPRatPA",L);
						end
					else
						return CalcStat("MitLightPRatPCap",L);
					end
				elseif SN > "PNTMPARMOURPENT" then
					if SN < "RESISTPRATP" then
						if SN < "PROGBRESIST" then
							if SN < "PNTMPRESIST" then
								if SN == "PNTMPBPE" then
									return 42/1200;
								else
									return 0;
								end
							elseif SN > "PNTMPRESIST" then
								if SN > "PROGBARMOUR" then
									if SN == "PROGBBPE" then
										return CalcStat("BRatProgB",L,"BRatStandard");
									else
										return 0;
									end
								elseif SN == "PROGBARMOUR" then
									return CalcStat("BRatProgB",L,"BRatMitMedium");
								else
									return 0;
								end
							else
								return 36/1200;
							end
						elseif SN > "PROGBRESIST" then
							if SN < "PROGEXTCOMLOWRAW" then
								if SN == "PROGEXTCOMHIGHRAW" then
									if Lm <= 121 then
										return ExpFmod(N,121,20,L,nil);
									elseif Lm <= 125 then
										return ExpFmod(CalcStat("ProgExtComHighRaw",121,N),122,5.5,L,nil);
									elseif Lm <= 126 then
										return ExpFmod(CalcStat("ProgExtComHighRaw",125,N),126,20,L,nil);
									elseif Lm <= 130 then
										return ExpFmod(CalcStat("ProgExtComHighRaw",126,N),127,5.5,L,nil);
									elseif Lm <= 131 then
										return ExpFmod(CalcStat("ProgExtComHighRaw",130,N),131,20,L,nil);
									elseif Lm <= 140 then
										return ExpFmod(CalcStat("ProgExtComHighRaw",131,N),132,5.5,L,nil);
									else
										return CalcStat("ProgExtComHighRaw",140,N);
									end
								else
									return 0;
								end
							elseif SN > "PROGEXTCOMLOWRAW" then
								if SN > "RESISTPBONUS" then
									if SN == "RESISTPPRAT" then
										return CalcRatAB(CalcStat("ResistPRatPA",L),CalcStat("ResistPRatPB",L),CalcStat("ResistPRatPCapR",L),N);
									else
										return 0;
									end
								elseif SN == "RESISTPBONUS" then
									return 0;
								else
									return 0;
								end
							else
								if Lm <= 116 then
									return ExpFmod(N,116,20,L,nil);
								elseif Lm <= 120 then
									return ExpFmod(CalcStat("ProgExtComLowRaw",116,N),117,5.5,L,nil);
								else
									return CalcStat("ProgExtComHighRaw",L,CalcStat("ProgExtComLowRaw",120,N));
								end
							end
						else
							return CalcStat("BRatProgB",L,"BRatExtra");
						end
					elseif SN > "RESISTPRATP" then
						if SN < "RESISTPRATPCAPR" then
							if SN < "RESISTPRATPB" then
								if SN == "RESISTPRATPA" then
									return 150;
								else
									return 0;
								end
							elseif SN > "RESISTPRATPB" then
								if SN > "RESISTPRATPC" then
									if SN == "RESISTPRATPCAP" then
										return 50;
									else
										return 0;
									end
								elseif SN == "RESISTPRATPC" then
									return 0.5;
								else
									return 0;
								end
							else
								return CalcStat("BRatRounded",L,"BRatExtra");
							end
						elseif SN > "RESISTPRATPCAPR" then
							if SN < "RUNEKEEPERCDHASPOWER" then
								if SN > "RESISTT" then
									if SN == "RUNEKEEPERCDARMOURTYPE" then
										return 1;
									else
										return 0;
									end
								elseif SN == "RESISTT" then
									return StatLinInter("PntMPResist","TraitProg","ProgBResist","",L,N,0);
								else
									return 0;
								end
							elseif SN > "RUNEKEEPERCDHASPOWER" then
								if SN > "T2PENARMOUR" then
									if SN == "T2PENBPE" then
										if Lm <= 115 then
											return (-40)*L;
										else
											return CalcStat("ProgExtComLowRaw",L,CalcStat("T2PenBPE",115));
										end
									else
										return 0;
									end
								elseif SN == "T2PENARMOUR" then
									return CalcStat("T2penMit",L);
								else
									return 0;
								end
							else
								return 1;
							end
						else
							return CalcStat("ResistPRatPB",L)*CalcStat("ResistPRatPC",L);
						end
					else
						return CalcPercAB(CalcStat("ResistPRatPA",L),CalcStat("ResistPRatPB",L),CalcStat("ResistPRatPCap",L),N);
					end
				else
					return 72/1200;
				end
			elseif SN > "T2PENMIT" then
				if SN < "TACMITLPRATPA" then
					if SN < "TACMITHPBONUS" then
						if SN < "TACDMGPRATPA" then
							if SN < "TACDMGPBONUS" then
								if SN == "T2PENRESIST" then
									if Lm <= 115 then
										return (-90)*L;
									else
										return CalcStat("ProgExtComLowRaw",L,CalcStat("T2PenResist",115));
									end
								else
									return 0;
								end
							elseif SN > "TACDMGPBONUS" then
								if SN > "TACDMGPPRAT" then
									if SN == "TACDMGPRATP" then
										return CalcStat("OutDmgPRatP",L,N);
									else
										return 0;
									end
								elseif SN == "TACDMGPPRAT" then
									return CalcStat("OutDmgPPRat",L,N);
								else
									return 0;
								end
							else
								return CalcStat("OutDmgPBonus",L);
							end
						elseif SN > "TACDMGPRATPA" then
							if SN < "TACDMGPRATPC" then
								if SN == "TACDMGPRATPB" then
									return CalcStat("OutDmgPRatPB",L);
								else
									return 0;
								end
							elseif SN > "TACDMGPRATPC" then
								if SN > "TACDMGPRATPCAP" then
									if SN == "TACDMGPRATPCAPR" then
										return CalcStat("OutDmgPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "TACDMGPRATPCAP" then
									return CalcStat("OutDmgPRatPCap",L);
								else
									return 0;
								end
							else
								return CalcStat("OutDmgPRatPC",L);
							end
						else
							return CalcStat("OutDmgPRatPA",L);
						end
					elseif SN > "TACMITHPBONUS" then
						if SN < "TACMITHPRATPC" then
							if SN < "TACMITHPRATP" then
								if SN == "TACMITHPPRAT" then
									return CalcStat("MitHeavyPPRat",L,N);
								else
									return 0;
								end
							elseif SN > "TACMITHPRATP" then
								if SN > "TACMITHPRATPA" then
									if SN == "TACMITHPRATPB" then
										return CalcStat("MitHeavyPRatPB",L);
									else
										return 0;
									end
								elseif SN == "TACMITHPRATPA" then
									return CalcStat("MitHeavyPRatPA",L);
								else
									return 0;
								end
							else
								return CalcStat("MitHeavyPRatP",L,N);
							end
						elseif SN > "TACMITHPRATPC" then
							if SN < "TACMITLPBONUS" then
								if SN > "TACMITHPRATPCAP" then
									if SN == "TACMITHPRATPCAPR" then
										return CalcStat("MitHeavyPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "TACMITHPRATPCAP" then
									return CalcStat("MitHeavyPRatPCap",L);
								else
									return 0;
								end
							elseif SN > "TACMITLPBONUS" then
								if SN > "TACMITLPPRAT" then
									if SN == "TACMITLPRATP" then
										return CalcStat("MitLightPRatP",L,N);
									else
										return 0;
									end
								elseif SN == "TACMITLPPRAT" then
									return CalcStat("MitLightPPRat",L,N);
								else
									return 0;
								end
							else
								return CalcStat("MitLightPBonus",L);
							end
						else
							return CalcStat("MitHeavyPRatPC",L);
						end
					else
						return CalcStat("MitHeavyPBonus",L);
					end
				elseif SN > "TACMITLPRATPA" then
					if SN < "TACMITMPRATPC" then
						if SN < "TACMITMPBONUS" then
							if SN < "TACMITLPRATPC" then
								if SN == "TACMITLPRATPB" then
									return CalcStat("MitLightPRatPB",L);
								else
									return 0;
								end
							elseif SN > "TACMITLPRATPC" then
								if SN > "TACMITLPRATPCAP" then
									if SN == "TACMITLPRATPCAPR" then
										return CalcStat("MitLightPRatPCapR",L);
									else
										return 0;
									end
								elseif SN == "TACMITLPRATPCAP" then
									return CalcStat("MitLightPRatPCap",L);
								else
									return 0;
								end
							else
								return CalcStat("MitLightPRatPC",L);
							end
						elseif SN > "TACMITMPBONUS" then
							if SN < "TACMITMPRATP" then
								if SN == "TACMITMPPRAT" then
									return CalcStat("MitMediumPPRat",L,N);
								else
									return 0;
								end
							elseif SN > "TACMITMPRATP" then
								if SN > "TACMITMPRATPA" then
									if SN == "TACMITMPRATPB" then
										return CalcStat("MitMediumPRatPB",L);
									else
										return 0;
									end
								elseif SN == "TACMITMPRATPA" then
									return CalcStat("MitMediumPRatPA",L);
								else
									return 0;
								end
							else
								return CalcStat("MitMediumPRatP",L,N);
							end
						else
							return CalcStat("MitMediumPBonus",L);
						end
					elseif SN > "TACMITMPRATPC" then
						if SN < "TPENCHOICE" then
							if SN < "TACMITMPRATPCAPR" then
								if SN == "TACMITMPRATPCAP" then
									return CalcStat("MitMediumPRatPCap",L);
								else
									return 0;
								end
							elseif SN > "TACMITMPRATPCAPR" then
								if SN > "TPENARMOUR" then
									if SN == "TPENBPE" then
										return -CalcStat("BPET",L,CalcStat("TpenChoice",N));
									else
										return 0;
									end
								elseif SN == "TPENARMOUR" then
									return -CalcStat("ArmourPenT",L,CalcStat("TpenChoice",N));
								else
									return 0;
								end
							else
								return CalcStat("MitMediumPRatPCapR",L);
							end
						elseif SN > "TPENCHOICE" then
							if SN < "WARDENCDARMOURTYPE" then
								if SN > "TPENRESIST" then
									if SN == "TRAITPROG" then
										return {{1,25,50,60,65,75,85,95,100,105,115,120,130,131,140,141},{1,25,50,60,65,75,85,95,100,105,115,120,130,131,140,141}};
									else
										return 0;
									end
								elseif SN == "TPENRESIST" then
									return -CalcStat("ResistT",L,CalcStat("TpenChoice",N)*2);
								else
									return 0;
								end
							elseif SN > "WARDENCDARMOURTYPE" then
								if SN > "WARDENCDCANBLOCK" then
									if SN == "WARDENCDHASPOWER" then
										return 1;
									else
										return 0;
									end
								elseif SN == "WARDENCDCANBLOCK" then
									return 1;
								else
									return 0;
								end
							else
								return 2;
							end
						else
							return DataTableValue({0,1,2},L);
						end
					else
						return CalcStat("MitMediumPRatPC",L);
					end
				else
					return CalcStat("MitLightPRatPA",L);
				end
			else
				if Lm <= 115 then
					return RoundDblDown(L*13.5)*-5;
				else
					return CalcStat("ProgExtComLowRaw",L,CalcStat("T2PenMit",115));
				end
			end
		else
			return CalcStat("MitHeavyPRatPB",L);
		end
	else
		return CalcStat("BPEPRatP",L,N);
	end

end

-- Support functions for CalcStat. These consist of implementations of more complex calculation types, decode methods for parameter "C" and rounding/min/max/compare functions for floating point numbers.

-- ****************** Calculation Type support functions ******************

-- DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD
-- DataTableValue: Takes a value from an array table.

function DataTableValue(vDataArray, dIndex)

	local lIndex = RoundDbl(dIndex);
	if lIndex <= 1 then
		return vDataArray[1];
	elseif lIndex > #vDataArray then
		return vDataArray[#vDataArray];
	else
		return vDataArray[lIndex];
	end

end

-- EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE
-- ExpFmod: Exponential function based on percentage.
-- Common percentage values are around ~5.5% for between levels and ~20% jumps between level segments.

function ExpFmod(dVal, dLstart, dPlvl, dLvl, vDec, vAdd)

	local dRng = dLvl-dLstart+1;
	if dRng <= DblCalcDev then
		return dVal;
	else
		local dFac = 1+dPlvl/100;
		local dAdd;
		if vAdd == nil then
			dAdd = 0;
		else
			dAdd = vAdd;
		end
		if vDec == nil then
			local dFacExp = dFac^dRng;
			return dVal*dFacExp+dAdd*((dFacExp-1)/(dFac-1));
		else
			local dResult = dVal;
			local dLm = dLstart-DblCalcDev;
			while dLm <= dLvl do
				dResult = RoundDbl(dResult*dFac+dAdd,vDec);
				dLm = dLm+1;
			end
			return dResult;
		end
	end

end

-- PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP
-- CalcPercAB: Calculates the percentage out of a rating based on the AB formula.

function CalcPercAB(dA, dB, dPCap, dR)

	if dR <= DblCalcDev then
		return 0.0;
	else
		local dResult = dA/(1+dB/dR);
		if dResult >= dPCap-DblCalcDev then
			return dPCap;
		else
			return dResult;
		end
	end

end

-- RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
-- CalcRatAB: Calculates the rating out of a percentage based on the AB formula.

function CalcRatAB(dA, dB, dCapR, dP)

	if dP <= DblCalcDev then
		return 0.0;
	else
		local dResult = dB/(dA/dP-1);
		if dResult >= dCapR-DblCalcDev then
			return dCapR;
		else
			return dResult;
		end
	end

end

-- SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
-- StatLinInter: (Normalized) Stat Linear Interpolating

function StatLinInter(sPntMP, sProgScheme, sProgBase, sAdj, dLvl, vNorC, vRoundType)

	-- parameter processing
	local dN = 1;
	local sC = "";
	if type(vNorC) == "number" then
		dN = vNorC;
	elseif type(vNorC) == "string" then
		sC = vNorC;
	end

	local dRoundType = 0;
	if type(vRoundType) == "number" then
		dRoundType = vRoundType;
	end
	
	local dProgScheme = CalcStat(sProgScheme,dLvl);
	if type(dProgScheme) ~= "table" then
		return 0.0;
	end

	-- find level interval
	local dLvlMinus = dLvl-DblCalcDev;
	local iPointIndexHigh = 2;
	local iPointIndexMax = #dProgScheme[1];
	while iPointIndexHigh < iPointIndexMax do
		if dLvlMinus <= dProgScheme[1][iPointIndexHigh] then
			break;
		end
		iPointIndexHigh = iPointIndexHigh+1;
	end
	local iPointIndexLow = iPointIndexHigh-1;
		
	local dAccessLvlLow = dProgScheme[1][iPointIndexLow];
	local dAccessLvlHigh = dProgScheme[1][iPointIndexHigh];
	local dBaseLvlLow = dProgScheme[2][iPointIndexLow];
	local dBaseLvlHigh = dProgScheme[2][iPointIndexHigh];
	
	-- get values from base progression
	local dValLow = CalcStat(sProgBase,dBaseLvlLow,sC);
	local dValHigh = CalcStat(sProgBase,dBaseLvlHigh,sC);

	-- graph point multiplications
	if type(sPntMP) == "string" and sPntMP ~= "" then
		dValLow = dValLow*CalcStat(sPntMP,dAccessLvlLow,sC);
		dValHigh = dValHigh*CalcStat(sPntMP,dAccessLvlHigh,sC);
	end
	if type(sAdj) == "string" and sAdj ~= "" then
		dValLow = dValLow*CalcStat(sAdj,dAccessLvlLow,sC);
		dValHigh = dValHigh*CalcStat(sAdj,dAccessLvlHigh,sC);
	end
	dValLow = dValLow*dN;
	dValHigh = dValHigh*dN;

	-- graph point roundings
	if dRoundType == 0 then
		dValLow = RoundDblLotro(dValLow);
		dValHigh = RoundDblLotro(dValHigh);
	elseif dRoundType == 1 then
		if -1000.0 <= dValLow and dValLow <= 1000.0 then
			dValLow = RoundDblUp(dValLow,2);
		else
			dValLow = RoundDblLotro(dValLow);
		end
		if -1000.0 <= dValHigh and dValHigh <= 1000.0 then
			dValHigh = RoundDblUp(dValHigh,2);
		else
			dValHigh = RoundDblLotro(dValHigh);
		end
	elseif dRoundType == 2 then
		dValLow = RoundDblLotro(dValLow);
		if dValLow == -1 then
			dValLow = -2;
		end
		dValHigh = RoundDblLotro(dValHigh);
		if dValHigh == -1 then
			dValHigh = -2;
		end
	end

	-- return interpolated value from the calculated graph points
	return LinFmod(1,dValLow,dValHigh,dAccessLvlLow,dAccessLvlHigh,dLvl);

end

-- TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT
-- LinFmod: Linear line function between 2 points with some optional modifications.
-- Connects point (dLstart,dVal*dFstart) with (dLend,dVal*dFend).
-- Usually used with dVal=1 and dFstart/dFend containing unrelated points or dVal=# and dFstart/dFend containing multiplier factors.
-- Modification for in-between points on the line: rounding.

function LinFmod(dVal, dFstart, dFend, dLstart, dLend, dLvl, vDec)

	if dLstart-DblCalcDev <= dLvl and dLvl <= dLstart+DblCalcDev then
		return dVal*dFstart;
	elseif dLend-DblCalcDev <= dLvl and dLvl <= dLend+DblCalcDev then
		return dVal*dFend;
	elseif dLstart == dLend then
		return 0.0;
	elseif vDec == nil then
		return dVal*(dFstart*(dLend-dLvl)+(dLvl-dLstart)*dFend)/(dLend-dLstart);
	else
		return RoundDbl(dVal*(dFstart*(dLend-dLvl)+(dLvl-dLstart)*dFend)/(dLend-dLstart),vDec);
	end

end

-- ****************** Parameter "C" decode support functions ******************

-- ArmCodeIndex: returns a specified index from an Armour Code.
-- sACode string:
-- 1st position: H=heavy, M=medium, L=light
-- 2nd position: H=head, S=shoulders, CL=cloak/back, C=chest, G=gloves, L=leggings, B=boots, Sh=shield
-- 3rd position: W=white/common, Y=yellow/uncommon, P=purple/rare, T=teal/blue/incomparable, G=gold/legendary/epic
-- Note: no such thing exists as a heavy, medium or light cloak, so no H/M/L in cloak codes (cloaks go automatically in the M class since U23, although historically this was L)

function ArmCodeIndex(sACode, iI)

	if sACode == nil or iI == nil then
		return 0;
	end
	if type(sACode) ~= "string" or type(iI) ~= "number" then
		return 0;
	end

	local sArmCode = string.match(sACode,"(%a+)");
	if not sArmCode then
		return 0;
	end
	sArmCode = string.upper(sArmCode).."    ";
	
	local sArmCat = string.sub(sArmCode,1,1);
	local sArmType = string.sub(sArmCode,2,2);
	local sArmCol = string.sub(sArmCode,3,3);

	if sArmType == "S" and sArmCol == "H" then
		sArmType = "SH";
		sArmCol = string.sub(sArmCode,4,4);
	elseif sArmCat == "C" and sArmType == "L" then
		sArmCat = "M";
		sArmType = "CL";
	elseif sArmType then
		sArmType = " "..sArmType;
	end
	
	local result = 0;
	if iI == 1 then
		if sArmCat then
			result = string.find("HML",sArmCat);
		end
	elseif iI == 2 then
		if sArmType then
			result = string.find(" H SCL C G L BSH",sArmType);
			if result then
				result = (result+1)/2;
			end
		end
	elseif iI == 3 then
		if sArmCol then
			result = string.find("WYPTG",sArmCol);
		end
	end

	if result then
		return result;
	else
		return 0;
	end
	
end

-- QualityCodeIndex: returns a quality index from a Quality Code.
-- sQCode string: W=white/common, Y=yellow/uncommon, P=purple/rare, T=teal/blue/incomparable, G=gold/legendary/epic

function QualityCodeIndex(sQCode)

	if sQCode == nil then
		return 0;
	end
	if type(sQCode) ~= "string" then
		return 0;
	end

	local sQtyCode = string.match(sQCode,"(%a+)");
	if not sQtyCode then
		return 0;
	end
	sQtyCode = string.upper(sQtyCode).." ";

	local sQtyCol = string.sub(sQtyCode,1,1);
	
	local result = 0;
	if sQtyCol then
		result = string.find("WYPTG",sQtyCol);
	end
	
	if result then
		return result;
	else
		return 0;
	end
	
end

-- WpnCodeIndex: returns a specified index from a Weapon Code.
-- sWCode string:
-- 1st position: H=heavy, L=light
-- 2nd position: O=one-handed, T=two-handed, B=bow
-- 3rd position: W=white/common, Y=yellow/uncommon, P=purple/rare, T=teal/blue/incomparable, G=gold/legendary/epic

function WpnCodeIndex(sWCode, iI)

	if sWCode == nil or iI == nil then
		return 0;
	end
	if type(sWCode) ~= "string" or type(iI) ~= "number" then
		return 0;
	end

	local sWpnCode = string.match(sWCode,"(%a+)");
	if not sWpnCode then
		return 0;
	end
	sWpnCode = string.upper(sWpnCode).."   ";
	
	local sWpnCat = string.sub(sWpnCode,1,1);
	local sWpnType = string.sub(sWpnCode,2,2);
	local sWpnCol = string.sub(sWpnCode,3,3);

	local result = 0;
	if iI == 1 then
		if sWpnCat then
			result = string.find("HL",sWpnCat);
		end
	elseif iI == 2 then
		if sWpnType then
			result = string.find("OTB",sWpnType);
		end
	elseif iI == 3 then
		if sWpnCol then
			result = string.find("WYPTG",sWpnCol);
		end
	end

	if result then
		return result;
	else
		return 0;
	end
	
end

-- RomanRankDecode: converts a string with a Roman number in characters, to an integer number.
-- used for Legendary Item Title calculation.

local RomanCharsToValues = {["M"]=1000,["CM"]=900,["D"]=500,["CD"]=400,["C"]=100,["XC"]=90,["L"]=50,["XL"]=40,["X"]=10,["IX"]=9,["V"]=5,["IV"]=4,["I"]=1};

function RomanRankDecode(sNumber)

	if sNumber ~= nil and type(sNumber) == "string" and sNumber ~= "" then
		for sRomanRankChar, iRomanRankValue in pairs(RomanCharsToValues) do
			if string.sub(string.upper(sNumber),1,string.len(sRomanRankChar)) == sRomanRankChar then
				return iRomanRankValue+RomanRankDecode(string.sub(sNumber,string.len(sRomanRankChar)+1));
			end
		end
	end
	
	return 0;
	
end

-- ReverseCalc: tries to calculate back a calculation result to the original (integer) level.
-- Does not support N.

function ReverseCalc(sStat, dNum)

	if string.len(string.match(sStat,"([-%w]+)")) > 0 then
		local dNumMinus = dNum-DblCalcDev;
		local dNumPlus = dNum+DblCalcDev;
		local minlvl = 1;
		local maxlvl = 549;
		local devlvl = 3;
	
		local left = minlvl-1;
		local right = maxlvl;
		local middle;
		
		local count = minlvl;

		while right > left+1 and count <= maxlvl do
			count = count+1;
			middle = math.floor((left+right)/2);
			if CalcStat(sStat,middle) >= dNumMinus then
				right = middle;
			else
				left = middle;
			end
		end

		local mintest = right-devlvl;
		if mintest < minlvl then
			mintest = minlvl;
		end
		local maxtest = right+devlvl;
		if maxtest > maxlvl then
			maxtest = maxlvl;
		end

		local dFound;
	
		-- we check nearby in case the progression is not completely ascending/sorted.
		for test = mintest, maxtest do
			dFound = CalcStat(sStat,test);
			if dNumMinus <= dFound and dFound <= dNumPlus then
				return test;
			end
		end
	end

	return 0;

end

-- ****************** Misc. floating point support functions ******************

-- Misc. functions for floating point rounding.
-- 2nd parameter is number of decimals.

function RoundDbl(dNum, vDec)

	local dCorrection = 0.5+DblCalcDev;
	local iSign = 1;
	if dNum < 0 then
		iSign = -1;
	end
	if vDec == nil or (-DblCalcDev <= vDec and vDec <= DblCalcDev) then
		return iSign*math.floor(iSign*dNum+dCorrection);
	else
		local dFactor = 10^vDec;
		return iSign*math.floor(iSign*dNum*dFactor+dCorrection)/dFactor;
	end
	
end

function RoundDblDown(dNum, vDec)

	local dCorrection = DblCalcDev;
	local iSign = 1;
	if dNum < 0 then
		iSign = -1;
	end
	if vDec == nil or (-DblCalcDev <= vDec and vDec <= DblCalcDev) then
		return iSign*math.floor(iSign*dNum+dCorrection);
	else
		local dFactor = 10^vDec;
		return iSign*math.floor(iSign*dNum*dFactor+dCorrection)/dFactor;
	end
	
end

function RoundDblUp(dNum, vDec)

	local dCorrection = 1-DblCalcDev;
	local iSign = 1;
	if dNum < 0 then
		iSign = -1;
	end
	if vDec == nil or (-DblCalcDev <= vDec and vDec <= DblCalcDev) then
		return iSign*math.floor(iSign*dNum+dCorrection);
	else
		local dFactor = 10^vDec;
		return iSign*math.floor(iSign*dNum*dFactor+dCorrection)/dFactor;
	end
	
end

function RoundDblLotro(dNum)

	local dCorrection = 1-DblCalcDev;
	local iSign = 1;
	if dNum < 0 then
		iSign = -1;
	end

	local iNumCeiled = math.floor(iSign*dNum+dCorrection);
	if iNumCeiled <= 1000 then
		return iSign*iNumCeiled;
	end

	local iFactor = 1;
	local iTestNum = math.floor(iNumCeiled/1000);
	while iTestNum > 0 do
		iTestNum = math.floor(iTestNum/10);
		iFactor = iFactor*10;
	end

	return iSign*math.floor(iNumCeiled/iFactor+dCorrection)*iFactor;

end

-- to be used by other modules
p.DblCalcDev = DblCalcDev;
p.CalcStat = CalcStat;

-- ******************************* End CalcStat *******************************
